"""Relocation types for i386.

Reference: https://github.com/hjl-tools/x86-psABI/wiki/intel386-psABI-1.1.pdf page 36
"""

from __future__ import annotations

from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericPCRelativeAddendReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
    RelocGOTMixin,
)


class R_386_32(GenericAbsoluteAddendReloc):
    """
    Value: 1
    Field: word32
    Calculation: S + A
    """


class R_386_PC32(GenericPCRelativeAddendReloc):
    """
    Value: 2
    Field: word32
    Calculation: S + A - P
    """


class R_386_COPY(GenericCopyReloc):
    """
    Value: 5
    Field:
    Calculation:
    """


class R_386_GLOB_DAT(GenericJumpslotReloc):
    """
    Value: 6
    Field: word32
    Calculation: S
    """


class R_386_JMP_SLOT(GenericJumpslotReloc):
    """
    Value: 7
    Field: word32
    Calculation: S
    """


class R_386_RELATIVE(GenericRelativeReloc):
    """
    Value: 8
    Field: word32
    Calculation: B + A
    """


class R_386_IRELATIVE(GenericIRelativeReloc):
    """
    Value: 42
    Field: word32
    Calculation: indirect (B + A)
    """


class R_386_TLS_DTPMOD32(GenericTLSModIdReloc):
    """
    Value: 35
    Field: word32
    Calculation:
    """


class R_386_TLS_TPOFF(GenericTLSOffsetReloc):
    """
    Value: 14
    Field: word32
    Calculation:
    """


class R_386_TLS_DTPOFF32(GenericTLSDoffsetReloc):
    """
    Value: 36
    Field: word32
    Calculation:
    """


class R_386_PLT32(GenericPCRelativeAddendReloc):
    """
    Value: 4
    Field: word32
    Calculation: L + A - P
    """


class R_386_GOTPC(GenericPCRelativeAddendReloc, RelocGOTMixin):
    """
    Value: 10
    Field: word32
    Calculation: GOT + A - P
    """


relocation_table_i386 = {
    1: R_386_32,
    2: R_386_PC32,
    # 3: R_386_GOT32,
    4: R_386_PLT32,
    5: R_386_COPY,
    6: R_386_GLOB_DAT,
    7: R_386_JMP_SLOT,
    8: R_386_RELATIVE,
    # 9: R_386_GOTOFF,
    10: R_386_GOTPC,
    14: R_386_TLS_TPOFF,
    # 15: R_386_TLS_IE,
    # 16: R_386_TLS_GOTIE,
    # 17: R_386_TLS_LE,
    # 18: R_386_TLS_GD,
    # 19: R_386_TLS_LDM,
    # 20: R_386_16,
    # 21: R_386_PC16,
    # 22: R_386_8,
    # 23: R_386_PC8,
    # 24: R_386_TLS_GD_32,
    # 25: R_386_TLS_GD_PUSH,
    # 26: R_386_TLS_GD_CALL,
    # 27: R_386_TLS_GD_POP,
    # 28: R_386_TLS_LDM_32,
    # 29: R_386_TLS_LDM_PUSH,
    # 30: R_386_TLS_LDM_CALL,
    # 31: R_386_TLS_LDM_POP,
    # 32: R_386_TLS_LDO_32,
    # 33: R_386_TLS_IE_32,
    # 34: R_386_TLS_LE_32,
    35: R_386_TLS_DTPMOD32,
    36: R_386_TLS_DTPOFF32,
    # 37: R_386_TLS_TPOFF32,
    # 38: R_386_SIZE32,
    # 39: R_386_TLS_GOTDESC,
    # 40: R_386_TLS_DESC_CALL,
    # 41: R_386_TLS_DESC,
    42: R_386_IRELATIVE,
    # 43: R_386_GOT32X,
}

__all__ = ("relocation_table_i386",)



from __future__ import annotations


class AddressTranslator:
    __slots__ = (
        "_rva",
        "_owner",
    )

    """
    Mediates address translations between typed addresses such as RAW, RVA, LVA, MVA and VA
    including address owner and its state (linked or mapped)

    Semantics::

        owner - object associated with the address
            (any object class based on `cle.Backend`)
        owner mapping state - sparse object can be either mapped or not
            (actual object's image base VA to be considered valid)
        RAW - offset (index) inside a file stream
        VA  - address inside process flat virtual memory space
        RVA - address relative to the object's segment base
            (segment base normalized virtual address)
        LVA - linked VA (linker)
        MVA - mapped VA (loader)
    """

    def __init__(self, rva, owner):
        """
        :param rva: virtual address relative to owner's object image base
        :type rva: int
        :param owner: The object owner address relates to
        :type owner: cle.Backend
        """
        self._rva, self._owner = rva, owner

    @classmethod
    def from_lva(cls, lva, owner):
        """
        Loads address translator with LVA
        """
        return cls(lva - owner.linked_base, owner)

    @classmethod
    def from_mva(cls, mva, owner):
        """
        Loads address translator with MVA
        """
        return cls(mva - owner.mapped_base, owner)

    @classmethod
    def from_rva(cls, rva, owner):
        """
        Loads address translator with RVA
        """
        return cls(rva, owner)

    @classmethod
    def from_raw(cls, raw, owner):
        """
        Loads address translator with RAW address
        """
        return cls(owner.offset_to_addr(raw) - (owner.mapped_base if owner._is_mapped else owner.linked_base), owner)

    from_linked_va = from_lva
    from_va = from_mapped_va = from_mva
    from_relative_va = from_rva

    def to_lva(self):
        """
        VA -> LVA
        :rtype: int
        """
        return self._rva + self._owner.linked_base

    def to_mva(self):
        """
        RVA -> MVA
        :rtype: int
        """
        return self._rva + self._owner.mapped_base

    def to_rva(self):
        """
        RVA -> RVA
        :rtype: int
        """
        return self._rva

    def to_raw(self):
        """
        RVA -> RAW
        :rtype: int
        """
        return self._owner.addr_to_offset(
            self._rva + (self._owner.mapped_base if self._owner._is_mapped else self._owner.linked_base)
        )

    to_linked_va = to_lva
    to_va = to_mapped_va = to_mva
    to_relative_va = to_rva


AT = AddressTranslator



from __future__ import annotations

import logging
import os

from cle.errors import CLEFileNotFoundError
from cle.utils import get_text_offset

log = logging.getLogger(name=__name__)


def convert_info_sharedlibrary(fname):
    """
    Convert a dump from gdb's ``info sharedlibrary`` command to a set of options that can be passed to CLE to replicate
    the address space from the gdb session

    :param fname:   The name of a file containing the dump
    :returns:       A dict appropriate to be passed as ``**kwargs`` for ``angr.Project`` or ``cle.Loader``
    """
    return _parse_gdb_map(fname, True)


def convert_info_proc_maps(fname):
    """
    Convert a dump from gdb's ``info proc maps`` command to a set of options that can be passed to CLE to replicate
    the address space from the gdb session

    :param fname:   The name of a file containing the dump
    :returns:       A dict appropriate to be passed as ``**kwargs`` for ``angr.Project`` or ``cle.Loader``
    """
    return _parse_gdb_map(fname, False)


def _parse_gdb_map(gdb_map, gdb_fix):
    if not os.path.isfile(gdb_map):
        raise CLEFileNotFoundError(f"gdb mapping file {gdb_map} does not exist")
    with open(gdb_map) as f:
        data = f.readlines()

    gmap = {}
    for line in data:
        if line in ("\n", "\r\n"):
            continue
        line_items = line.split()
        # Get rid of all metadata, just extract lines containing addresses
        if "0x" not in line_items[0]:
            continue
        if line_items[-1].startswith("["):
            continue
        try:
            int(line_items[-1], 16)
        except ValueError:
            pass
        else:
            continue

        addr, objfile = int(line_items[0], 16), line_items[-1].strip()

        # Get the smallest address of each libs' mappings
        try:
            gmap[objfile] = min(gmap[objfile], addr)
        except KeyError:
            gmap[objfile] = addr

    # Find lib names
    # libnames = filter(lambda n: '.so' in n, gmap.keys())

    # Find base addr for each lib (each lib is mapped to several segments,
    # we take the segment that is loaded at the smallest address).
    lib_opts = {}
    main_opts = {}
    force_load_libs = []
    smallest_addr = min(gmap.values())

    for lib, addr in gmap.items():
        if addr == smallest_addr and not gdb_fix:
            # this is the main binary
            opts = main_opts
        else:
            # if not os.path.exists(lib):
            #    lib = _simple_search(lib)
            force_load_libs.append(lib)
            opts = {}
            lib_opts[lib] = opts

        # address of .text -> base address of the library
        if gdb_fix:
            found = _simple_search(lib)
            if found:
                addr = addr - get_text_offset(found)

        log.info("gdb_plugin: mapped %s to %#x", lib, addr)
        opts["base_addr"] = addr

    return {"force_load_libs": force_load_libs, "main_opts": main_opts, "lib_opts": lib_opts}


GDB_SEARCH_PATH = ["/lib", "/usr/lib"]


def _simple_search(libname):
    dirs = list(GDB_SEARCH_PATH)
    while dirs:
        dirname = dirs.pop(0)
        try:
            for name in os.listdir(dirname):
                if name in (".", ".."):
                    continue
                full = os.path.join(dirname, name)
                if os.path.isdir(full):
                    if full.count("/") < 12:  # don't go too deep
                        dirs.append(full)
                if os.path.isfile(full) and name == libname:
                    return full
        except OSError:
            pass
    return libname



from __future__ import annotations

from .elf_tls import ELFThreadManager
from .elfcore_tls import ELFCoreThreadManager
from .minidump_tls import MinidumpThreadManager
from .pe_tls import PEThreadManager
from .tls_object import InternalTLSRelocation, ThreadManager, TLSObject

__all__ = [
    "ThreadManager",
    "InternalTLSRelocation",
    "TLSObject",
    "ELFThreadManager",
    "ELFCoreThreadManager",
    "PEThreadManager",
    "MinidumpThreadManager",
]



from __future__ import annotations

import os.path

from cle.address_translator import AT

from .subprogram import Subprogram
from .variable import Variable


class CompilationUnit:
    """
    CompilationUnit for DWARF
    See http://dwarfstd.org/doc/DWARF5.pdf page 60
    """

    def __init__(self, name, comp_dir, low_pc, high_pc, language, elf_object):
        self.name = name
        self.comp_dir = comp_dir
        self.file_path = os.path.join(self.comp_dir, self.name)
        self.low_pc = low_pc
        self.high_pc = high_pc
        self.language = language
        self.functions: dict[int, Subprogram] = {}
        self.global_variables: list[Variable] = []
        self._elf_object = elf_object

    @property
    def min_addr(self):
        return AT.from_rva(self.low_pc, self._elf_object).to_mva()

    @property
    def max_addr(self):
        return AT.from_rva(self.high_pc, self._elf_object).to_mva()
